from django.db import models


class Site(models.Model):
    def __unicode__(self):
        return u'Site {0}'.format(self.id)


class Customer(models.Model):
    site = models.ForeignKey('Site', related_name='customers')

    def __unicode__(self):
        return u'Customer {0} - Site {1}'.format(self.id, self.site.id)


class CustomerUser(models.Model):
    customer = models.ForeignKey('Customer', related_name='users')

    def __unicode__(self):
        return u'User {0} - Customer {1}'.format(self.id, self.customer.id)



import collections


# Base

class BaseInstanceDict(collections.Mapping):
    def __init__(self, descriptor, instance):
        self.descriptor = descriptor
        self.instance = instance

    def __repr__(self):
        return repr(dict(self))

    def __iter__(self):
        return self.descriptor.model_configs.iterkeys()

    def __len__(self):
        return len(self.descriptor.model_configs)

    def __getitem__(self, item):
        model_config = self.descriptor.model_configs[item]
        section = model_config.section
        kwargs = model_config.get_kwargs(self.instance)
        return self.get_value(section, kwargs)

    def get_value(self, section, kwargs):
        raise NotImplementedError


class DescriptorBase(object):
    instance_dict = None

    def __init__(self, model_config):
        self.check()
        self.model_configs = {}
        self.add_model_config(model_config)

    def check(self):
        if self.instance_dict is None:
            raise NotImplementedError
        if not issubclass(self.instance_dict, BaseInstanceDict):
            raise ValueError

    def add_model_config(self, model_config):
        self.model_configs[model_config.section.name] = model_config

    def remove_model_config(self, model_config):
        del self.model_configs[model_config.section.name]

    def __get__(self, instance, type=None):
        if instance is None:
            return self
        return self.instance_dict(self, instance)


class ManagerBase(object):
    descriptor = None

    def __init__(self, model_config):
        self.check()
        self.model_config = model_config

    def check(self):
        if self.descriptor is None:
            raise NotImplementedError
        if not issubclass(self.descriptor, DescriptorBase):
            raise ValueError

    def contribute_to_class(self, cls, name, **kwargs):
        if not hasattr(cls, name):
            setattr(cls, name, self.descriptor(self.model_config))
        else:
            getattr(cls, name).add_model_config(self.model_config)


# Preferences

class PreferencesInstanceDict(BaseInstanceDict):
    def get_value(self, section, kwargs):
        return section.get_preferences(**kwargs)


class PreferencesDescriptor(DescriptorBase):
    instance_dict = PreferencesInstanceDict


class PreferencesManager(ManagerBase):
    descriptor = PreferencesDescriptor


# Permissions

class PermissionsInstanceDict(BaseInstanceDict):
    def get_value(self, section, kwargs):
        return section.get_permissions(**kwargs)


class PermissionsDescriptor(DescriptorBase):
    instance_dict = PermissionsInstanceDict


class PermissionsManager(ManagerBase):
    descriptor = PermissionsDescriptor



from prefs_n_perms.client import db
from prefs_n_perms.permissions import Permissions
from prefs_n_perms.preferences import Preferences
from prefs_n_perms.settings import preference_settings


class Section(object):
    prefix = preference_settings.SECTIONS_PREFIX

    def __init__(self, name, **kwargs):
        self.name = name

    def __str__(self):
        return self.name

    @property
    def section_key(self):
        return ':'.join((self.prefix, self.name))

    def exists(self):
        return self.section_key in db

    @property
    def has_preferences(self):
        return Preferences(self).exists()

    @property
    def has_permissions(self):
        return Permissions(self).exists()

    def get_preferences(self, **kwargs):
        return Preferences(self, **kwargs)

    def get_permissions(self, **kwargs):
        return Permissions(self, **kwargs)

    @property
    def tiers(self):
        return db.List(self.section_key)

    @tiers.setter
    def tiers(self, value):
        if not isinstance(value, (tuple, list)):
            raise ValueError
        if self.exists():
            db.api.delete(self.section_key)
        db.List(self.section_key, value)



from redis.connection import DefaultParser
from prefs_n_perms import utils
from prefs_n_perms.settings import preference_settings as settings


class ConnectionFactory(object):

    # Store connection pool by cache backend options.
    #
    # _pools is a process-global, as otherwise _pools is cleared every time
    # ConnectionFactory is instiated, as Django creates new cache client
    # (DefaultClient) instance for every request.

    _pools = {}

    def __init__(self, options):
        pool_cls_path = options.get("CONNECTION_POOL_CLASS",
                                    "redis.connection.ConnectionPool")
        self.pool_cls = utils.load_class(pool_cls_path)
        self.pool_cls_kwargs = options.get("CONNECTION_POOL_KWARGS", {})

        redis_client_cls_path = options.get("REDIS_CLIENT_CLASS",
                                            "redis.client.StrictRedis")
        self.redis_client_cls = utils.load_class(redis_client_cls_path)
        self.redis_client_cls_kwargs = options.get("REDIS_CLIENT_KWARGS", {})

        self.options = options

    def make_connection_params(self, url):
        """
        Given a main connection parameters, build a complete
        dict of connection parameters.
        """

        kwargs = {
            "url": url,
            "parser_class": self.get_parser_cls(),
        }

        password = self.options.get("PASSWORD", None)
        if password:
            kwargs["password"] = password

        socket_timeout = self.options.get("SOCKET_TIMEOUT", None)
        if socket_timeout:
            assert isinstance(socket_timeout, (int, float)), \
                "Socket timeout should be float or integer"
            kwargs["socket_timeout"] = socket_timeout

        socket_connect_timeout = self.options.get("SOCKET_CONNECT_TIMEOUT", None)
        if socket_connect_timeout:
            assert isinstance(socket_connect_timeout, (int, float)), \
                "Socket connect timeout should be float or integer"
            kwargs["socket_connect_timeout"] = socket_connect_timeout

        return kwargs

    def connect(self, url):
        """
        Given a basic connection parameters,
        return a new connection.
        """
        params = self.make_connection_params(url)
        connection = self.get_connection(params)
        return connection

    def get_connection(self, params):
        """
        Given a now preformated params, return a
        new connection.
        The default implementation uses a cached pools
        for create new connection.
        """
        pool = self.get_or_create_connection_pool(params)
        return self.redis_client_cls(connection_pool=pool, **self.redis_client_cls_kwargs)

    def get_parser_cls(self):
        cls = self.options.get("PARSER_CLASS", None)
        if cls is None:
            return DefaultParser
        return utils.load_class(cls)

    def get_or_create_connection_pool(self, params):
        """
        Given a connection parameters and return a new
        or cached connection pool for them.
        Reimplement this method if you want distinct
        connection pool instance caching behavior.
        """
        key = params["url"]
        if key not in self._pools:
            self._pools[key] = self.get_connection_pool(params)
        return self._pools[key]

    def get_connection_pool(self, params):
        """
        Given a connection parameters, return a new
        connection pool for them.
        Overwrite this method if you want a custom
        behavior on creating connection pool.
        """
        cp_params = dict(params)
        cp_params.update(self.pool_cls_kwargs)
        pool = self.pool_cls.from_url(**cp_params)

        if pool.connection_kwargs.get("password", None) is None:
            pool.connection_kwargs["password"] = params.get("password", None)
            pool.reset()

        return pool


def get_connection_factory(path=None, options=None):
    options = options or settings
    if path is None:
        path = options.get("CONNECTION_FACTORY",
                           "prefs_n_perms.pool.ConnectionFactory")

    cls = utils.load_class(path)
    return cls(options)



from importlib import import_module

from prefs_n_perms.exceptions import PrefsNPermsException


def load_class(path):
    """
    Loads class from path.
    """

    mod_name, klass_name = path.rsplit('.', 1)

    try:
        mod = import_module(mod_name)
    except AttributeError as e:
        raise PrefsNPermsException('Error importing {0}: "{1}"'.format(mod_name, e))

    try:
        klass = getattr(mod, klass_name)
    except AttributeError:
        raise PrefsNPermsException('Module "{0}" does not define a "{1}" class'.format(mod_name, klass_name))

    return klass



from prefs_n_perms import pool
from redish.client import Client
from redish.serialization import Plain
from prefs_n_perms.settings import preference_settings as settings


def get_client(url=None):
    connection_factory = pool.get_connection_factory()
    return connection_factory.connect(url or settings.REDIS_URL)


redis = get_client()


class RedishClient(Client):
    serializer = Plain()

    def __init__(self, *args, **kwargs):
        self.api = redis


db = RedishClient()



from django.apps import AppConfig
from prefs_n_perms.registries import autodiscover


class PrefsNPermsConfig(AppConfig):
    name = 'prefs_n_perms'
    verbose_name = 'Preferences and Permissions'

    def ready(self):
        # Loop through each app and look for preferences registry
        autodiscover()



from collections import namedtuple, OrderedDict
import collections
from prefs_n_perms.base import PrefsNPermsBase
from prefs_n_perms.client import db
from prefs_n_perms.decorators import protect_writable, cached_property
from prefs_n_perms.settings import preference_settings

PermissionSet = namedtuple('PermissionSet', ['allowed', 'blocked'])


class Permissions(PrefsNPermsBase, collections.MutableSet):
    prefix = preference_settings.PERMISSIONS_PREFIX
    global_tier = 'available'

    def __repr__(self):
        return repr(self.all)

    def __contains__(self, item):
        return item in self.all

    def __iter__(self):
        return iter(self.all)

    def __len__(self):
        return len(self.all)

    def get_db_obj(self, key):
        return db.Set(key)

    def get_global_db_obj(self):
        return db.SortedSet(self.global_key)

    def add_global_db_obj(self):
        pass

    def add_tier_db_obj(self, tier, key):
        self.tiers[tier] = PermissionSet(self.get_db_obj(key + ':allowed'),
                                         self.get_db_obj(key + ':blocked'))

    @protect_writable
    def _add(self, db_obj, *vals):
        if len(vals) == 1 and isinstance(vals[0], (list, tuple, set)):
            vals = list(vals[0])
        db_obj.client.sadd(db_obj.name, *vals)

    @protect_writable
    def _rem(self, db_obj, *vals):
        if len(vals) == 1 and isinstance(vals[0], (list, tuple, set)):
            vals = list(vals[0])
        db_obj.client.srem(db_obj.name, *vals)

    @property
    def all_available(self):
        return self.get_global_db_obj()

    @property
    def available(self):
        blocked = set()
        [blocked.update(b) for tier, (_, b) in self.tiers.iteritems() if tier != self.last_tier]
        return [p for p in self.all_available if p not in blocked]

    def clear_available(self):
        try:
            db.api.delete(self.global_key)
        except KeyError:
            pass

    def update_available(self, perms):
        if perms:
            self.clear_available()
            perms = reduce(tuple.__add__, enumerate(perms))
            db.api.zadd(self.global_key, *perms)

    def get_allowed_for(self, tier):
        return self.get_tier(tier).allowed

    def is_allowed_for(self, tier, name):
        return name in self.get_allowed_for(tier)

    def add_allowed_for(self, tier, *perms):
        allowed = self.get_allowed_for(tier)
        self._add(allowed, *perms)

    def remove_allowed_for(self, tier, *perms):
        allowed = self.get_allowed_for(tier)
        self._rem(allowed, *perms)

    def get_blocked_for(self, tier):
        return self.get_tier(tier).blocked

    def is_blocked_for(self, tier, name):
        return name in self.get_blocked_for(tier)

    def add_blocked_for(self, tier, *perms):
        blocked = self.get_blocked_for(tier)
        self._add(blocked, *perms)

    def remove_blocked_for(self, tier, *perms):
        blocked = self.get_blocked_for(tier)
        self._rem(blocked, *perms)

    def add(self, *perms):
        if perms:
            self.add_allowed_for(self.last_tier, *perms)

    def remove(self, *perms):
        if perms:
            self.remove_allowed_for(self.last_tier, *perms)
    discard = remove

    def block(self, *perms):
        if perms:
            self.add_blocked_for(self.last_tier, *perms)

    def unblock(self, *perms):
        if perms:
            self.remove_blocked_for(self.last_tier, *perms)

    @protect_writable
    def clear(self):
        try:
            tier = self.get_tier(self.last_tier)
            db.api.delete(tier.allowed.name)
            db.api.delete(tier.blocked.name)
        except KeyError:
            pass

    @property
    def all_tiers(self):
        perms = OrderedDict()
        for tier, (allowed, blocked) in self.tiers.iteritems():
            perms[tier] = {
                'allowed': set(allowed),
                'blocked': set(blocked),
            }
        return perms

    @cached_property
    def all(self):
        perms = PermissionSet(set(), set())
        for allowed, blocked in self.tiers.itervalues():
            perms.allowed.update(allowed)
            perms.blocked.update(blocked)
        return perms.allowed - perms.blocked

    @property
    def inherited(self):
        perms = PermissionSet(set(), set())
        for tier, (allowed, blocked) in self.tiers.iteritems():
            if tier != self.last_tier:
                perms.allowed.update(allowed)
                perms.blocked.update(blocked)
        return perms.allowed - perms.blocked

    def has_perm(self, name):
        return name in self.all



from django.test import TestCase
from test.test_app.models import Site, Customer, CustomerUser


class TestSections(TestCase):

    def setUp(self):
        super(TestCase, self).setUp()

class TestPreferences(TestCase):

    def setUp(self):
        super(TestCase, self).setUp()
        self.site = Site.objects.create()
        self.customer = Customer.objects.create(site=self.site)
        self.customer_user = CustomerUser.objects.create(customer=self.customer)

    def test_default_preferences(self):
        self.assertEqual(self.site.preferences, self.customer.preferences)
        self.assertEqual(self.site.preferences, self.customer_user.preferences)

class TestPermissions(TestCase):

    def setUp(self):
        super(TestCase, self).setUp()



from prefs_n_perms.managers import PreferencesManager, PermissionsManager
from prefs_n_perms.sections import Section


class BaseConfig(object):
    pass


class SectionConfig(BaseConfig):
    tiers = ()
    default_preferences = {}
    available_permissions = ()

    def __init__(self, name):
        self.name = name
        self.section = Section(name)
        self.initialize()

    def initialize(self):
        # set up initial preferences and permissions
        if not self.section.exists() and self.tiers:
            self.section.tiers = self.tiers
        if not self.section.has_preferences and self.default_preferences:
            self.section.get_preferences().update_global(self.default_preferences)
        if not self.section.has_permissions and self.available_permissions:
            self.section.get_permissions().update_available(self.available_permissions)


class ModelConfig(BaseConfig):
    model = None

    def __init__(self, section_config, tier):
        if self.model is None:
            raise NotImplementedError
        self.section_config = section_config
        self.section = section_config.section
        self.tier = tier
        self.attach_to_model()

    def __del__(self):
        self.detach_from_model()

    def get_kwargs(self, obj):
        raise NotImplementedError

    def attach_to_model(self):
        if self.section.has_preferences:
            self.model.add_to_class('preferences', PreferencesManager(self))
        if self.section.has_permissions:
            self.model.add_to_class('permissions', PermissionsManager(self))

    def detach_from_model(self):
        if hasattr(self.model, 'preferences'):
            self.model.preferences.remove_model_config(self)
        if hasattr(self.model, 'permissions'):
            self.model.permissions.remove_model_config(self)



from prefs_n_perms import config, section_registry, model_registry
from test.test_app.models import Site, Customer, CustomerUser


class CustomerConfig(config.SectionConfig):
    tiers = ('site', 'customer', 'customer_user')
    default_preferences = {
        'theme': 'orange',
        'show_help_on_login': True,
    }
    available_permissions = (
        'can_invite_user',
        'can_disable_user',
        'can_access_billing',
        'can_reset_password',
    )

class SiteModelConfig(config.ModelConfig):
    model = Site

    def get_kwargs(self, obj):
        return {
            'site': obj.id,
        }

class CustomerModelConfig(config.ModelConfig):
    model = Customer

    def get_kwargs(self, obj):
        return {
            'site': obj.site.id,
            'customer': obj.id,
        }

class CustomerUserModelConfig(config.ModelConfig):
    model = CustomerUser

    def get_kwargs(self, obj):
        return {
            'site': obj.customer.site.id,
            'customer': obj.customer.id,
            'customer_user': obj.id,
        }


section_registry.register('customers', CustomerConfig)
model_registry.register('customers', 'site', SiteModelConfig)
model_registry.register('customers', 'customer', CustomerModelConfig)
model_registry.register('customers', 'customer_user', CustomerUserModelConfig)



from collections import defaultdict
import inspect
from django.conf import settings
from importlib import import_module
from prefs_n_perms.config import SectionConfig, ModelConfig
from prefs_n_perms.exceptions import SectionAlreadyRegisteredException, SectionNotRegisteredException, \
    TierAlreadyRegisteredException
from prefs_n_perms.sections import Section
from prefs_n_perms.settings import preference_settings


class SectionRegistry(dict):
    def register(self, section, config):
        if not issubclass(config, SectionConfig) or not inspect.isclass(config):
            raise ValueError
        if section in self:
            raise SectionAlreadyRegisteredException

        config = config(section)
        self[section] = config

        return config

    def unregister(self, section):
        del self[section]


class ModelRegistry(defaultdict):
    def __init__(self, **kwargs):
        super(ModelRegistry, self).__init__(dict, **kwargs)

    def register(self, section, tier, config):
        if not issubclass(config, ModelConfig) or not inspect.isclass(config):
            raise ValueError
        if section not in section_registry:
            raise SectionNotRegisteredException
        if tier in self[section]:
            raise TierAlreadyRegisteredException

        section_config = section_registry[section]
        config = config(section_config, tier)
        self[section][tier] = config

        return config

    def unregister(self, section, tier):
        del self[section][tier]


def autodiscover():
    """
    Populate the registry by iterating through every section declared in :py:const:`settings.INSTALLED_APPS`.
    """
    for app in settings.INSTALLED_APPS:
        package = '{0}.{1}'.format(app, preference_settings.REGISTRY_MODULE)
        try:
            import_module(package)
        except ImportError:
            pass


section_registry = SectionRegistry()
model_registry = ModelRegistry()



from collections import OrderedDict
import collections
from prefs_n_perms.base import PrefsNPermsBase
from prefs_n_perms.client import db
from prefs_n_perms.decorators import protect_writable, cached_property
from prefs_n_perms.settings import preference_settings


class Preferences(PrefsNPermsBase, collections.MutableMapping):
    prefix = preference_settings.PREFERENCES_PREFIX

    def __repr__(self):
        return repr(self.all)

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        return self.set(key, value)

    def __delitem__(self, key):
        return self.remove(key)

    def __iter__(self):
        return iter(self.all.keys())

    def __len__(self):
        return len(self.all)

    def get_db_obj(self, key):
        return db.Dict(key)

    def get_for(self, tier, name):
        return self.get_tier(tier).get(name)

    @protect_writable
    def update_tier(self, tier, new_prefs):
        prefs = self.get_tier(tier)
        prefs.update(new_prefs)

    @protect_writable
    def set_for(self, tier, name, value):
        prefs = self.get_tier(tier)
        prefs[name] = value

    @protect_writable
    def remove_for(self, tier, *name):
        prefs = self.get_tier(tier)
        prefs.client.hdel(prefs.name, *name)

    def update_global(self, new_prefs):
        self.update_tier(self.global_tier, new_prefs)

    def set_global(self, name, value):
        self.set_for(self.global_tier, name, value)

    def remove_global(self, *name):
        self.remove_for(self.global_tier, *name)

    def update(*args, **kwargs):
        if len(args) > 2:
            raise TypeError("update() takes at most 2 positional "
                            "arguments ({} given)".format(len(args)))
        elif not args:
            raise TypeError("update() takes at least 1 argument (0 given)")
        self = args[0]
        other = {}
        other.update(args[1] if len(args) >= 2 else {})
        other.update(kwargs)
        self.update_tier(self.last_tier, other)

    def set(self, name, value):
        self.set_for(self.last_tier, name, value)

    def remove(self, *name):
        self.remove_for(self.last_tier, *name)

    @cached_property
    def all(self):
        prefs = {}
        for tier in self.tiers.itervalues():
            prefs.update(tier)
        return prefs

    def get(self, name, default=None):
        pref = default
        for tier in self.tiers.itervalues():
            pref = tier.get(name, pref)
        return pref

    def all_tiers(self):
        prefs = OrderedDict()
        for tier, db_obj in self.tiers.iteritems():
            prefs[tier] = dict(db_obj)
        return prefs



from os.path import dirname, abspath, join
from django.conf.global_settings import *

BASE_DIR = dirname(dirname(abspath(__file__)))


DEBUG = False # will be False anyway by DjangoTestRunner.
TEMPLATE_DEBUG = False
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': join(BASE_DIR, 'db.sqlite3'),
    }
}
MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

TEMPLATE_CONTEXT_PROCESSORS = TEMPLATE_CONTEXT_PROCESSORS + (
    'django.core.context_processors.request',
    # 'dynamic_preferences.processors.global_preferences',
    # 'dynamic_preferences.processors.user_preferences',
)
INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.sites',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.admin',

    'test.test_app',
    'prefs_n_perms',
)

ROOT_URLCONF = 'test.urls'

PREFS_N_PERMS = {
    'REDIS_URL': 'redis://localhost:6379/0',
}

SITE_ID = 1
STATIC_URL = "/static/"
TEMPLATE_LOADERS = (
    'django.template.loaders.filesystem.Loader',
    'django.template.loaders.app_directories.Loader',
)

PREFS_N_PERMS_USE_TEST_PREFERENCES=True
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'simple': {
            'format': '%(levelname)s %(message)s'
        },
    },
    'handlers': {
        'console':{
            'level':'DEBUG',
            'class':'logging.StreamHandler',
            'formatter': 'simple'
        },
    },
    'loggers': {
        'django.request':{
            'handlers': ['console'],
            'propagate': True,
            'level': 'DEBUG',
        },
    },
}
TESTING=True
TEST_RUNNER='django_nose.NoseTestSuiteRunner'
NOSE_ARGS = [
    '--with-coverage',
    '--cover-package=prefs_n_perms',
]


from collections import OrderedDict
from prefs_n_perms.client import db
from prefs_n_perms.decorators import protect_writable
from prefs_n_perms.exceptions import ExtraInstancesException, MissingInstancesException, GlobalVariableException
from prefs_n_perms.settings import preference_settings


class PrefsNPermsBase(object):
    prefix = preference_settings.BASE_PREFIX
    global_tier = 'global'
    cached_objects = ('all',)

    def __init__(self, section, require_all=False, read_only=False, **kwargs):
        self.section = section
        self.require_all = require_all
        self.read_only = read_only
        self.instance_keys = self.get_instance_keys(kwargs)
        self.tiers = OrderedDict()
        self.add_tiers()

    def key_for(self, entity):
        return ':'.join((self.prefix, self.section.name, entity))

    def get_instance_keys(self, kwargs):
        """
        Loop through (tier, id) in kwargs, check if ID is a number
        and return a dict((tier, db_key), ...)
        """
        return (dict((k, self.key_for('.'.join((k, str(v)))))
                     for k, v in kwargs.iteritems()
                     if str(v).isdigit())
                if kwargs else {})

    @property
    def global_key(self):
        return self.key_for(self.global_tier)

    def exists(self):
        return self.global_key in db

    def clear_cache(self):
        for obj in self.cached_objects:
            if hasattr(self, obj):
                delattr(self, obj)

    def add_tiers(self):
        self.add_global_db_obj()
        self.validate()

        instance_tiers = self.instance_keys.keys()
        missing_tiers = []

        for tier in self.section.tiers:
            if tier in self.instance_keys:
                self.add_tier_db_obj(tier, self.instance_keys[tier])
                instance_tiers.remove(tier)
            elif len(instance_tiers):
                missing_tiers.append(tier)
            else:
                break

        if missing_tiers:
            raise MissingInstancesException(missing_tiers)

    def validate(self):
        section_tiers = set(self.section.tiers)
        instance_tiers = set(self.instance_keys)

        # If tiers that are not in the section are in the instance keys, raise exception
        if instance_tiers > section_tiers:
            raise ExtraInstancesException(instance_tiers - section_tiers)

        # If require_all is true and instance_tiers does not contain all section_tiers, raise exception
        if self.require_all and instance_tiers < section_tiers:
            raise MissingInstancesException(section_tiers - instance_tiers)

    def get_db_obj(self, key):
        return db.get(key)

    def get_global_db_obj(self):
        return self.get_db_obj(self.global_key)

    def add_global_db_obj(self):
        self.tiers[self.global_tier] = self.get_global_db_obj()

    def add_tier_db_obj(self, tier, key):
        self.tiers[tier] = self.get_db_obj(key)

    @property
    def last_tier(self):
        tier = self.tiers.keys()[-1]
        if tier == self.global_tier:
            raise GlobalVariableException
        return tier

    def get_tier(self, tier):
        return self.tiers[tier]

    def get_last_tier(self):
        return self.get_tier(self.last_tier)

    @protect_writable
    def clear(self):
        try:
            prefs = self.get_tier(self.last_tier)
            db.api.delete(prefs.name)
        except KeyError:
            pass





class PrefsNPermsException(Exception):
    detail_default = 'An exception occurred with Preferences and Permissions'

    def __init__(self, detail=None):
        if detail is not None:
            self.detail = str(detail)
        else:
            self.detail = str(self.detail_default)

    def __str__(self):
        return self.detail


class ReadOnlyException(PrefsNPermsException):
    detail_default = 'Cannot write to read-only settings'


class InvalidInstancesException(PrefsNPermsException):
    detail_default = 'Invalid instance(s): {instances}'

    def __init__(self, instances, detail=None):
        if detail is not None:
            self.detail = str(detail)
        else:
            instances = ', '.join(list(instances))
            self.detail = self.detail_default.format(instances=instances)


class MissingInstancesException(InvalidInstancesException):
    default_detail = 'Missing instance(s): {instances}'


class ExtraInstancesException(InvalidInstancesException):
    default_detail = 'Extra instance(s): {instances}'


class GlobalVariableException(PrefsNPermsException):
    detail_default = 'Can not set global variable'


class SectionAlreadyRegisteredException(PrefsNPermsException):
    detail_default = 'This section has already been registered'


class SectionNotRegisteredException(PrefsNPermsException):
    detail_default = 'This section has not been registered'


class TierAlreadyRegisteredException(PrefsNPermsException):
    detail_default = 'This tier has already been registered'



import os


def _get_git_revision(path):
    revision_file = os.path.join(path, 'refs', 'heads', 'master')
    if not os.path.exists(revision_file):
        return None
    fh = open(revision_file, 'r')
    try:
        return fh.read().strip()[:7]
    finally:
        fh.close()


def get_revision():
    """
    :returns: Revision number of this branch/checkout, if available. None if
        no revision number can be determined.
    """
    package_dir = os.path.dirname(__file__)
    checkout_dir = os.path.normpath(os.path.join(package_dir, os.pardir))
    path = os.path.join(checkout_dir, '.git')
    if os.path.exists(path):
        return _get_git_revision(path)
    return None


def get_version():
    base = __version__
    if __build__:
        base = '%s (%s)' % (base, __build__)
    return base


__version__ = '0.2.5'
__build__ = get_revision()
VERSION = get_version()

default_app_config = 'prefs_n_perms.apps.PrefsNPermsConfig'
from prefs_n_perms.registries import section_registry, model_registry



from setuptools import setup, find_packages

setup(
    name='prefs-n-perms',
    version='0.2.5',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'hiredis',
        'redis',
        'redish',
    ],
    zip_safe=True,
    url='https://github.com/revpoint/prefs-n-perms',
    author='RevPoint Media LLC',
    author_email='jangl@revpointmedia.com',
    description='Manage preferences and permissions with hierarchy structure',
)



from django.core.management.base import BaseCommand
from prefs_n_perms import section_registry


class Command(BaseCommand):
    help = 'Restore all default preferences to Redis DB.'

    def handle(self, *args, **options):
        for section, config in section_registry.iteritems():
            self.stdout.write('Setting preferences for: %s' % section)
            if config.default_preferences:
                config.section.get_preferences().update_global(config.default_preferences)
            if config.available_permissions:
                config.section.get_permissions().update_available(config.available_permissions)
